import mock
import ssl

import ZEO.asyncio.mtacceptor
import ZEO.Exceptions
import ZEO.runzeo
import ZEO.StorageServer
import ZODB

from .base import get_admin
from .ownerstorage import OwnerStorage

class Acceptor(ZEO.asyncio.mtacceptor.Acceptor):

    def __init__(self, storage_server, addr, ssl):
        super(Acceptor, self).__init__(storage_server, addr, ssl)
        [self.cert_storage_id] = storage_server.storages # TCBOO
        storage = storage_server.storages[self.cert_storage_id]
        self.cert_db = ZODB.DB(storage)
        self.invalidate = self.cert_db._mvcc_storage.invalidate
        with self.cert_db.transaction() as conn:
            self.certs_oid = get_admin(conn).certs._p_oid

    @property
    def ssl_context(self):
        context = self.storage_server.create_ssl_context()
        with self.cert_db.transaction() as conn:
            certs = conn.get(self.certs_oid)
            context.load_verify_locations(cadata=certs.data)
        return context

    @ssl_context.setter
    def ssl_context(self, context):
        pass

class ZEOStorage(ZEO.StorageServer.ZEOStorage):

    user_id = None

    registered_methods = frozenset(
        ZEO.StorageServer.registered_methods | set(['get_root_id']))

    def register(self, storage_id, read_only, credentials=None):
        super(ZEOStorage, self).register(storage_id, read_only)

        der = self.connection.transport.get_extra_info(
            'ssl_object').getpeercert(1)

        with ZODB.DB(self.storage).transaction() as conn:
            admin = get_admin(conn)
            self.user_id = admin.uids[der]
            if self.user_id is None:
                # Nobody cert.
                if not credentials:
                    raise ZEO.Exceptions.AuthError()

            if credentials:
                user = admin.users_by_name[credentials['name']]
                if ((user.id != self.user_id and self.user_id is not None) or
                    not user.check_password(credentials['password'])
                    ):
                    raise ZEO.Exceptions.AuthError()
                self.user_id = user.id

        self.storage = OwnerStorage(self.storage, self.user_id)

    def setup_delegation(self):
        super(ZEOStorage, self).setup_delegation()
        self.connection.methods = self.registered_methods

    def get_root_id(self):
        return self.user_id

class StorageServer(ZEO.StorageServer.StorageServer):

    def invalidate(self, conn, storage_id, tid, invalidated=(), info=None):
        """ Internal: broadcast info and invalidations to clients. """

        # Same as in the base class
        if invalidated:
            invq = self.invq[storage_id]
            if len(invq) >= self.invq_bound:
                invq.pop()
            invq.insert(0, (tid, invalidated))

        for zs in self.zeo_storages_by_storage_id[storage_id]:
            if conn and zs.user_id == conn.user_id:
                connection = zs.connection
                if invalidated and zs is not conn:
                    # zs.client.invalidateTransaction(tid, invalidated)
                    connection.call_soon_threadsafe(
                        connection.async,
                        'invalidateTransaction', tid, invalidated)
                elif info is not None:
                    # zs.client.info(info)
                    connection.call_soon_threadsafe(
                        connection.async, 'info', info)

        # Update the cert storage db:
        if storage_id == self.acceptor.cert_storage_id:
            self.acceptor.invalidate(tid, invalidated)

    def create_client_handler(self):
        return ZEOStorage(self, self.read_only)


class ZeroDBOptions(ZEO.runzeo.ZEOOptions):

    def realize(self, *args):
        # Evil evil hack, but ZConfig schemas hurt and we need to
        # ovverride the way ssl configuration is handled. :(
        with mock.patch('ssl.create_default_context'):
            super(ZeroDBOptions, self).realize(*args)


class ZEOServer(ZEO.runzeo.ZEOServer):
    def create_server(self):
        storages = self.storages
        options = self.options
        self.server = StorageServer(
            options.address,
            storages,
            read_only=options.read_only,
            client_conflict_resolution=True,
            invalidation_queue_size=options.invalidation_queue_size,
            invalidation_age=options.invalidation_age,
            transaction_timeout=options.transaction_timeout,
            Acceptor=Acceptor,
            )

        # See evil evil mock hack above :(
        ssl_args, ssl_kw_args = options.ssl.load_cert_chain.call_args
        def create_ssl_context():
            context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            context.load_cert_chain(*ssl_args, **ssl_kw_args)
            context.verify_mode = ssl.CERT_REQUIRED
            return context

        self.server.create_ssl_context = create_ssl_context

    @classmethod
    def run(cls, args=None):
        options = ZeroDBOptions()
        options.realize(args)

        for o_storage in options.storages:
            if o_storage.config.pack_gc:
                logging.warn(
                    "Packing with GC and end-to-end encryption"
                    " removes all the data")
                logging.warn("Turining GC off!")
                o_storage.config.pack_gc = False

        s = cls(options)
        s.main()




from BTrees.Length import Length
from zope.index.text.lexicon import Lexicon as _Lexicon

from zerodb import trees
from zerodb.storage import parallel_traversal


def _text2list(text):
    # Helper: splitter input may be a string or a list of strings
    # Fixed from zope.index.text.lexicon
    try:
        text + u""
    except UnicodeDecodeError:
        try:
            return [text.decode("utf-8")]
        except UnicodeDecodeError:
            # It wasn't utf-8. But at least, it was text!
            return [text]
    except TypeError:
        return text
    else:
        return [text]


class Lexicon(_Lexicon):
    family = trees.family32  # In comparison with standard Lexicon, use bigger buckets

    def __init__(self, *pipeline):
        self._wids = self.family.OI.BTree()
        self._words = self.family.IO.BTree()
        self.wordCount = Length()
        self._pipeline = pipeline

    def sourceToWordIds(self, text):
        if text is None:
            text = ''
        last = _text2list(text)
        for element in self._pipeline:
            last = element.process(last)
        if not isinstance(self.wordCount, Length):
            # Make sure wordCount is overridden with a BTrees.Length.Length
            self.wordCount = Length(self.wordCount())
        # Strategically unload the length value so that we get the most
        # recent value written to the database to minimize conflicting wids
        # Because length is independent, this will load the most
        # recent value stored, regardless of whether MVCC is enabled
        self.wordCount._p_deactivate()
        parallel_traversal(self._wids, last)
        return list(map(self._getWordIdCreate, last))

    def termToWordIds(self, text):
        last = _text2list(text)
        for element in self._pipeline:
            last = element.process(last)
        wids = []
        if len(last) > 1:
            parallel_traversal(self._wids, last)
        for word in last:
            wids.append(self._wids.get(word, 0))
        return wids

    # XXX globToWordIds should pre-fetch prefixes in *range*



"""
Key derivation/extraction functions.

Functions have form of:

def security_function(username, password, key_file, cert_file, appname, key):
    ...
    return password, key

Where key is 256-bit symmetric key, and new_password could be hash

We use scrypt to derive the key on the client side
"""
import scrypt
import hashlib

scrypt_kw = dict(N=16384, r=8, p=1, buflen=32)


def key_from_password(username, password, key_file, cert_file, appname, key):
    """
    Key is derived from password using scrypt
    password is replaced with sha256(key) so that we cannot deduce key from
    auth info
    """
    salt = "|".join([username, appname, 'key'])
    key = scrypt.hash(password, salt, **scrypt_kw)
    # Calculate hash so that server cannot derive key from new_password
    # b'auth' is added so that we don't use hash(key) by mistake
    new_password = hashlib.sha256(key + b'auth').digest()
    return new_password, key


def key_from_cert(username, password, key_file, cert_file, appname, key):
    """
    Key is derived from SSL key_file.
    Password is securely hashed if given at all
    """
    with open(key_file) as f:
        key = hashlib.sha256(f.read().encode()).digest()

    if password is not None:
        salt = "|".join([username, appname, 'key'])
        password = scrypt.hash(password, salt, **scrypt_kw)
        password = hashlib.sha256(password + b'auth').digest()

    return password, key


def hash_password(username, password, key_file, cert_file, appname, key):
    """
    Password is hashed, encryption key is left untouched.
    Password is compatible with other functions
    """
    if password:
        salt = "|".join([username, appname, 'key'])
        password = scrypt.hash(password, salt, **scrypt_kw)
        password = hashlib.sha256(password + b'auth').digest()

    return password, key


def guess(username, password, key_file, cert_file, appname, key):
    """
    Guess which function to use from values arguments
    """
    if key is None:
        if password is not None:
            return key_from_password
        elif key_file is not None:
            return key_from_cert
        else:
            raise AttributeError("Not enough attributes to guess KDF")

    else:
        return hash_password



import itertools
import os
import ssl
import threading

import six
from six.moves import zip as izip
from Crypto import Random

import transaction
import ZODB
import ZODB.Connection

from zerodbext.catalog.query import optimize
from zerodb.collective.indexing.indexer import PortalCatalogProcessor
from zerodb.collective.indexing.interfaces import IIndexQueueProcessor
from zerodb.collective.indexing import subscribers
from zope import component

from zerodb import models
from zerodb.catalog.query import And, Eq
from zerodb.models.exceptions import ModelException
from zerodb.storage import client_storage
from zerodb.util.thread_watcher import ThreadWatcher
from zerodb.util.iter import DBList, DBListPrefetch, Sliceable

from zerodb.transform.encrypt_aes import AES256Encrypter, AES256EncrypterV0
from zerodb.transform import init_crypto
from zerodb.crypto import kdf


class AutoReindexQueueProcessor(PortalCatalogProcessor):
    def __init__(self, db, enabled=True):
        self.db = db
        self.enabled = enabled

    # execute reindex in before_commit hook when commit
    def reindex(self, obj, attributes=None):
        if self.enabled:
            self.db.reindex(obj, attributes)


class DbModel(object):
    """
    Class where model is combined with db.
    All functionality will actually reside here.
    Will contain indexes as well.
    """
    def __init__(self, db, model):
        """
        :param zerodb.DB db: Database to link model to
        :param model: Data model (subclass of zerodb.models.Model)
        """
        self._model = model
        self._db = db
        self._catalog_name = "catalog__" + model.__modelname__
        self._intid_name = "store__" + model.__modelname__
        if not transaction.manager._txn and \
                (self._intid_name not in db._root or self._catalog_name not in db._root):
            transaction.begin()
            commit = True
        else:
            commit = False

        if self._intid_name not in db._root:
            _objects = model.create_store()
            db._root[self._intid_name] = _objects

        if self._catalog_name not in db._root:
            _catalog = model.create_catalog()
            db._root[self._catalog_name] = _catalog

        if commit:
            transaction.commit()

    @property
    def _catalog(self):
        return self._db._root[self._catalog_name]

    @property
    def _objects(self):
        return self._db._root[self._intid_name]

    def __getitem__(self, uids):
        """
        DbModels (which we query) are accessed by using db as a dictionary

        :param int uids: object's uid or list of them
        :return: Persistent object(s)
        """
        if isinstance(uids, six.integer_types):
            obj = self._objects[uids]
            if not hasattr(obj, "_p_uid"):
                obj._p_uid = uids
            return obj

        elif isinstance(uids, (tuple, list, set)):
            objects = [self._objects[uid] for uid in uids]
            self._db._connection.prefetch(objects)
            for o, uid in izip(objects, uids):
                if not hasattr(o, "_p_uid"):
                    o._p_uid = uid
            return objects

        else:
            raise ModelException("Integer or list of integers is expected")

    def all_uids(self):
        for i in self._objects.tree:
            yield i

    def all(self):
        for i in self.all_uids():
            obj = self._objects[i]
            if not hasattr(obj, "_p_uid"):
                obj._p_uid = i
            yield obj

    def add(self, obj):
        """
        Add newly created a Model object to the database
        Stores *and* indexes it

        :param zerodb.models.Model obj: Object to add to the database
        :return: Added object's uid
        :rtype: int
        """
        assert obj.__class__ == self._model
        uid = self._objects.add(obj)
        self._catalog.index_doc(uid, obj)
        obj._p_uid = uid
        return uid

    def reindex_one(self, obj, attributes=None):
        """
        Reindex one object which is already in the database

        :param obj: Object to add to the database or its uid
        :type obj: zerodb.models.Model, int
        :param attributes: Attributes of obj to be reindex
        :type attributes: tuple, list
        """

        if isinstance(obj, six.integer_types):
            uid = obj
            obj = self._objects[uid]
        elif isinstance(obj, self._model):
            if hasattr(obj, "_p_uid"):
                uid = obj._p_uid
            else:
                raise ModelException("Object %s is not indexed" % obj)
        else:
            raise TypeError("Wrong type of argument passed: obj must be integer or model instance")

        if attributes is None:
            self._catalog.reindex_doc(uid, obj)
        elif isinstance(attributes, (tuple, list)):
            for attr in attributes:
                if attr in self._catalog:
                    self._catalog[attr].reindex_doc(uid, obj)
        else:
            raise TypeError("Wrong type of argument passed: attributes must be tuple or list")

    def reindex(self, obj):
        """
        Reindex one or multiple objects in the database

        :param obj: Object to add to the database or its uid, or list of objects or uids
        :type obj: zerodb.models.Model, int, list
        """
        if isinstance(obj, six.integer_types) or isinstance(obj, self._model):
            self.reindex_one(obj)
        elif isinstance(obj, (list, tuple, set, Sliceable)):
            for o in obj:
                self.reindex_one(o)
        else:
            raise TypeError("ZeroDB object, its uid or list of these should be passed")

    def remove(self, obj):
        """
        Remove existing object from the database + unindex it

        :param zerodb.models.Model obj: Object to add to the database
        """
        if isinstance(obj, six.integer_types):
            uid = obj
        elif hasattr(obj, "__iter__"):
            ctr = 0
            for i in list(obj):
                self.remove(i)
                ctr += 1
            return ctr
        else:
            assert obj.__class__ == self._model
            uid = obj._p_uid
        self._catalog.unindex_doc(uid)
        del self._objects[uid]
        return 1

    def query(self, queryobj=None, skip=None, limit=None, prefetch=True, **kw):
        """
        Smart proxy to catalog's query.
        One can add <field=...> keyword arguments to make queries where fields
        are equal to specified values

        :param zerodb.catalog.query.Query queryobj: Query which all sorts of
            logical, range queries etc
        :param int skip: Offset to start the result iteration from
        :param int limit: Limit number of results to this
        """
        # Catalog's query returns only integers
        # We must be smart here and return objects
        # But no, we must be even smarter and batch-preload objects
        # Most difficult part is preloading TreeSets for index when needed
        # (when we do complex queries which require composite index)
        # We also probably should do something like lazy query(...)[ofs:...]
        # if no limit, skip are used

        # Work needed on skip and limit because zope didn't well support them...
        skip = skip or 0
        if limit:
            kw["limit"] = skip + limit

        eq_args = []
        for k in list(kw.keys()):
            if k not in set(["sort_index", "sort_type", "reverse", "names", "limit"]):
                eq_args.append(Eq(k, kw.pop(k)))

        if queryobj:
            Q = optimize(optimize(queryobj) & And(*eq_args))
        else:
            Q = And(*eq_args)

        q = lambda: self._catalog.query(Q, **kw)

        if limit:
            _, q = q()
            # XXX islice -> [:]
            qids = list(itertools.islice(q, skip, skip + limit))
            objects = [self._objects[uid] for uid in qids]
            if objects and prefetch:
                self._db._connection.prefetch(objects)
            for obj, uid in izip(objects, qids):
                obj._p_uid = uid
            return objects

        else:
            db_list = DBListPrefetch if prefetch else DBList
            return db_list(q, self)

    def __len__(self):
        return len(self._objects)


class SubConnection(ZODB.Connection.Connection):

    @property
    def root(self):
        return ZODB.Connection.RootConvenience(self.get(self._db._root_oid))

    def setstate(self, obj):
        if hasattr(obj, "_p_uid"):
            uid = obj._p_uid
        else:
            uid = None
        super(SubConnection, self).setstate(obj)
        if uid is not None:
            obj._p_uid = uid


class SubDB(ZODB.DB):
    klass = SubConnection

    def __init__(self, storage, *a, **kw):
        super(SubDB, self).__init__(storage, *a, **kw)
        self._root_oid = storage.get_root_id()


class DB(object):
    """
    Database for this user. Everything is used through this class
    """

    encrypter = [AES256Encrypter, AES256EncrypterV0]
    appname = 'zerodb.com'
    compressor = None

    def __init__(self, sock, key=None, username=None, password=None,
                 cert_file=None, key_file=None, server_cert=None,
                 security=None,
                 debug=False, pool_timeout=3600, pool_size=7,
                 autoreindex=True, wait_timeout=30,
                 **kw):
        """
        :param str sock: UNIX (str) or TCP ((str, int)) socket
        :type sock: str or tuple
        :param str username: Username
        :param str password: Password
        :param bytes key: Encryption key

        :param cert_file: Client certificate for authentication (pem)
        :param key_file: Private key for that certificate (pem)
        :param server_cert: Server certitificate if not registered with CA

        :param function security: Key derivation function from
                                  zerodb.crypto.kdf

        :param bool debug: Whether to log debug messages
        """

        if (cert_file or key_file) and not (cert_file and key_file):
            raise TypeError("If you specify a cert file or a key file,"
                            " you must specify both.")

        ssl_context = make_ssl(cert_file, key_file, server_cert)

        if security is None:
            security = kdf.guess(
                    username, password, key_file, cert_file, self.appname, key)

        password, key = security(
                username, password, key_file, cert_file, self.appname, key)

        if password:
            credentials = dict(name=username, password=password)
        else:
            credentials = None

        if isinstance(sock, six.string_types):
            sock = str(sock)
        elif type(sock) in (list, tuple):
            assert len(sock) == 2
            sock = str(sock[0]), int(sock[1])

        self._autoreindex = autoreindex
        self._reindex_queue_processor = AutoReindexQueueProcessor(
            self, enabled=autoreindex)
        component.provideUtility(
            self._reindex_queue_processor,
            IIndexQueueProcessor,
            'zerodb-indexer')

        self._init_default_crypto(key=key)

        # Store all the arguments necessary for login in this instance
        self.__storage_kwargs = {
                "sock": sock,
                "ssl": ssl_context,
                "cache_size": 2 ** 30,
                "debug": debug,
                "wait_timeout": wait_timeout,
                "credentials": credentials,
            }

        self.__db_kwargs = {
                "pool_size": pool_size,
                "pool_timeout": pool_timeout,
                "cache_size": 1000000,
                "cache_size_bytes": 100 * 2 ** 20}
        self.__db_kwargs.update(kw)

        # For multi-threading
        self.__pid = os.getpid()

        self._init_db()
        self._models = {}

    @classmethod
    def _init_default_crypto(self, **kw):
        encrypters = self.encrypter
        if not isinstance(encrypters, (list, tuple)):
            encrypters = [self.encrypter]
        elif not encrypters:
            encrypters = []

        if encrypters:
            encrypters[0].register_class(default=True)
            for e in encrypters[1:]:
                e.register_class(default=False)
        if self.compressor:
            self.compressor.register(default=True)

        init_crypto(**kw)

    def _init_db(self):
        """We need this to be executed each time we are in a new process"""
        if self._autoreindex:
            subscribers.init()

        Random.atfork()

        self.__conn_refs = {}
        self.__thread_local = threading.local()
        self.__thread_watcher = ThreadWatcher()
        self._storage = client_storage(**self.__storage_kwargs)
        self._db = SubDB(self._storage, **self.__db_kwargs)
        self._conn_open()

    def _conn_open(self):
        """Opens db connection and registers a destuction callback"""

        self.__thread_local.conn = conn = self._db.open()

        def destructor(conn_id):
            conn = self._db.pool.all.data.get(conn_id, None)
            if conn:
                # Hack to make it not closing TM which is already removed
                conn.opened = 0
                conn.close()

        self.__thread_watcher.watch(destructor, id(conn))

    @property
    def _root(self):
        """Access database root for this user"""

        if os.getpid() != self.__pid:
            # If a new process spins up, we need to re-initialize everything
            self.__pid = os.getpid()
            self._init_db()
        else:
            # Open connections from the pool when new threads spin up
            # Should be closed when old thread-locals get garbage collected
            if not hasattr(self.__thread_local, "conn") or\
                    self.__thread_local.conn.opened is None:
                self._conn_open()

        return self.__thread_local.conn.root()

    @property
    def _connection(self):
        return self.__thread_local.conn

    def disconnect(self):
        if hasattr(self.__thread_local, "conn"):
            self.__thread_local.conn.close()

    def __getitem__(self, model):
        """
        DbModels (which we query) are accessed by using db as a dictionary

        :param model: Subclass of zerodb.models.Model to return or create db entry for
        :rtype: zerodb.db.DbModel
        """
        # TODO implement list of keys, writing to arbitrary (new) dbmodel (which is not defined)
        if not issubclass(model, models.Model):
            raise ModelException("Class <%s> is not a Model" % model.__name__)
        if model not in self._models:
            self._models[model] = DbModel(self, model)
        return self._models[model]

    def add(self, obj):
        """
        Add newly created a Model object to the database
        Stores *and* indexes it

        :param zerodb.models.Model obj: Object to add to the database
        :return: Added object's uid
        :rtype: int
        """
        if isinstance(obj, (list, set, tuple)):
            return [self[o.__class__].add(o) for o in obj]
        else:
            return self[obj.__class__].add(obj)

    def remove(self, obj):
        """
        Remove existing object from the database + unindex it

        :param zerodb.models.Model obj: Object to add to the database
        """
        if isinstance(obj, models.Model):
            self[obj.__class__].remove(obj)
            return 1
        elif hasattr(obj, "__iter__"):
            ctr = 0
            for o in obj:
                ctr += 1
                self[o.__class__].remove(o)
            return ctr
        else:
            raise ModelException("Class <%s> is not a Model or iterable" % obj.__class__.__name__)

    def reindex(self, obj, attributes=None):
        """
        Reindex one or multiple objects in the database

        :param obj: Object to add to the database or its uid, or list of objects or uids
        :type obj: zerodb.models.Model, list
        :param attributes: Attributes of obj to be reindex
        :type attributes: tuple, list
        """
        if isinstance(obj, models.Model):
            self[obj.__class__].reindex_one(obj, attributes)
        elif isinstance(obj, (list, tuple, set, Sliceable)):
            for o in obj:
                assert isinstance(o, models.Model)
                self[o.__class__].reindex_one(o, attributes)
        else:
            raise TypeError("ZeroDB object or list of these should be passed")

    def pack(self):
        """
        Remove old versions of objects
        """
        self._db.pack()

    def enableAutoReindex(self, enabled=True):
        """
        Enable or disable auto reindex
        """
        if enabled:
            subscribers.init()
        self._reindex_queue_processor.enabled = enabled


def make_ssl(cert_file=None, key_file=None, server_cert=None):
    ssl_context = ssl.create_default_context(
        ssl.Purpose.CLIENT_AUTH, cafile=server_cert)
    here = os.path.dirname(__file__)
    ssl_context.load_cert_chain(
        cert_file or os.path.join(here, 'permissions/nobody.pem'),
        key_file or os.path.join(here, 'permissions/nobody-key.pem'),
        )
    ssl_context.verify_mode = ssl.CERT_REQUIRED
    if server_cert is None:
        ssl_context.check_hostname = True
        ssl_context.load_default_certs()
    else:
        ssl_context.check_hostname = False
    return ssl_context



from zope.interface import Attribute, Interface


class ICompressor(Interface):
    """Compressing or decompressing data"""
    name = Attribute("Signature that object is compressed with this algorithm. Recorded as '.cname$'")
    _compress = Attribute("Low-level compress function")
    _decompress = Attribute("Low-level decompress function")

    def compress(data):
        """Compresses data"""

    def decompress(data):
        """Decompresses data"""

    def register(default):
        """Register utility"""


class IEncrypterClass(Interface):
    """Class which marks encrypting interface, not encrypting object"""


class IEncrypter(Interface):
    """Encrypting or decrypting data"""
    name = Attribute("Signature that object is encrypted with this algorithm. Recorded as '.ename$'")
    attributes = Attribute("List of attributes to consume from init")

    def encrypt(data):
        """Encrypts data"""

    def decrypt(data):
        """Decrypts data"""

    def _encrypt(data):
        """Low level encrypt interface"""

    def _decrypt(data):
        """Low level decrypt interface"""

    def _init_encryption(**kw):
        """Extra functions to init encryption"""



from zc.zlibstorage import ZlibStorage
from ZODB.utils import maxtid
import logging
import zope.component
import zope.interface
from zerodb.transform import encrypt, decrypt, compress, decompress
from zerodb.transform import get_encryption_signature
from zerodb.util import encode_hex
from zerodb.util.debug import debug_loads
import zerodb.transform.interfaces

_gsm = zope.component.getGlobalSiteManager()

class TransformingStorage(ZlibStorage):
    """
    Storage which can transform (encrypt and/or compress) data.
    """

    def __init__(self, base, *args, **kw):
        """
        :param base: Storage to transform
        :param bool debug: Output debug log messages
        """
        self.base = base

        self.debug = kw.pop("debug", False)
        if self.debug:
            self._debug_download_size = 0
            self._debug_download_count = 0

        for name in self.copied_methods:
            v = getattr(base, name, None)
            if v is not None:
                setattr(self, name, v)

        zope.interface.directlyProvides(self, zope.interface.providedBy(base))

        base.registerDB(self)

        self._transform = lambda data: encrypt(compress(data), no_cipher_name=True)
        self._transform_named = lambda data: encrypt(compress(data), no_cipher_name=False)
        self._untransform = lambda data: decompress(decrypt(data))

        self._root_oid = base.get_root_id()

        # configure default encrytion based on root encryption:
        data = base.loadBefore(self._root_oid, maxtid)[0]
        sig = get_encryption_signature(data)
        if sig:
            _gsm.registerUtility(
                _gsm.getUtility(
                    zerodb.transform.interfaces.IEncrypter,
                    sig.decode(),
                    )
                )

    def loadBefore(self, oid, tid):
        """Load last state for a given oid before a given tid

        :param str oid: Object ID
        :param str tid: Transaction timestamp
        :return: Object and its serial number and following serial number
        :rtype: tuple
        """
        if self.debug:
            if oid not in self._cache.current:
                in_cache = False
            else:
                in_cache = True

        data, serial, tend = self.base.loadBefore(oid, tid)
        out_data = self._untransform(data)

        if self.debug and not in_cache:
            logging.debug(
                "id:%s, type:%s, transform: %s->%s" % (
                    encode_hex(oid),
                    debug_loads(out_data),
                    len(data),
                    len(out_data),
                    ))
            self._debug_download_size += len(data)
            self._debug_download_count += 1

        return out_data, serial, tend

    def store(self, oid, serial, data, version, transaction):
        if oid == self._root_oid:
            _transform = self._transform_named
        else:
            _transform = self._transform
        return self.base.store(oid, serial, _transform(data), version,
                               transaction)



import six

import ZEO.ClientStorage
from itertools import chain
from persistent import Persistent

from . import transforming
import logging

# TODO when it comes to the point we need to,
# we'll have to configure which classes to use
# with Zope interfaces

class ClientStorage(ZEO.ClientStorage.ClientStorage):

    def get_root_id(self):
        return self._call('get_root_id')

def client_storage(sock, *args, **kw):
    """
    Storage client

    :param sock: UNIX or TCP socket
    :param cipher: Encryptor to use (see zerodb.crypto)
    :param bool debug: Output debug messages to the log
    :param transforming_storage: Wrapper-storage to use (TransformingStorage)
    :returns: Storage
    :rtype: TransformingStorage
    """
    if six.PY2 and isinstance(sock, unicode):
        sock = str(sock)
    TransformingStorage = kw.pop(
        'transforming_storage', transforming.TransformingStorage)
    debug = kw.pop("debug", False)
    return TransformingStorage(ClientStorage(sock, *args, **kw),
                               debug=debug)


def prefetch(objs):
    """
    Bulk-fetch ZODB objects
    """
    objs = [o for o in objs if getattr(o, "_p_oid", None)]
    if objs and objs[0]._p_jar:
        objs[0]._p_jar.prefetch(objs)


def prefetch_trees(trees, depth=10, bucket_types=(), shallow=True):
    """
    Bulk-fetch specified trees or buckets in logarithmic number of steps
    """
    trees = [t for t in trees
             if isinstance(t, Persistent) and hasattr(t, "_p_oid")]

    if len(trees) == 0 or depth == 0:
        # If someone inserts an infinite loop, we shouldn't go for that
        return

    prefetch(trees)

    if not bucket_types:
        bucket_types = tuple(set(chain(*[
            (type(t), t._bucket_type) for t in trees
            if hasattr(t, "_bucket_type")])))

    children = []
    for tree in trees:
        state = tree.__getstate__()
        if state:
            children += [o for o in state[0] if isinstance(o, bucket_types)]

    prefetch_trees(children, depth=(depth - 1), bucket_types=bucket_types)


def btree_state_search(state, key):
    """
    Search in raw state of BTree.
    state = ((branch, key, branch, key, ... , key, branch), 1st_bucket) or None
    """
    if not state:
        return -1, None

    state = state[0]

    # Based on python BTree logic
    lo = 0
    hi = (len(state) + 1) // 2
    i = hi // 2
    while i > lo:
        i_key = state[i * 2 - 1]
        if i_key < key:
            lo = i
        elif i_key > key:
            hi = i
        else:
            break
        i = (lo + hi) // 2
    return i, state[i * 2]


def parallel_traversal(trees, keys):
    """
    Traverse trees in parallel to fill up cache
    """
    if not isinstance(trees, (list, tuple)):
        to_fetch = [trees]
        trees = [trees] * len(keys)
    else:
        to_fetch = list(set(
            t for t in trees
            if isinstance(t, Persistent) and getattr(t, "_p_oid", None)
            ))

    prefetch(to_fetch)

    nxt_trees = []
    nxt_keys = []
    for key, tree in zip(keys, trees):
        if isinstance(tree, Persistent) and hasattr(tree, "_bucket_type"):
            _, nxt = btree_state_search(tree.__getstate__(), key)
            if isinstance(nxt, (type(tree), tree._bucket_type)):
                nxt_keys.append(key)
                nxt_trees.append(nxt)

    if nxt_keys:
        parallel_traversal(nxt_trees, nxt_keys)



from zope.component import getGlobalSiteManager, ComponentLookupError
from zope.interface import implementer
from .interfaces import IEncrypter, IEncrypterClass

_gsm = getGlobalSiteManager()


@implementer(IEncrypter)
class CommonEncrypter(object):
    name = b""
    attributes = ()

    def __init__(self, **kw):
        kwargs = {}
        for i in self.attributes:
            if i in kw:
                kwargs[i] = kw[i]
        self._signature = b".e%s$" % self.name
        self._init_encryption(**kwargs)
        self.register()

    def _init_encryption(self, **kw):
        pass

    def encrypt(self, data, no_cipher_name=False):
        if no_cipher_name:
            sig = b".e$"
        else:
            sig = self._signature
        return sig + self._encrypt(data)

    def decrypt(self, data):
        if data.startswith(self._signature):
            return self._decrypt(data[len(self._signature):])
        elif data.startswith(b".e$"):
            return self._decrypt(data[3:])
        else:
            return data

    def register(self):
        try:
            if _gsm.getUtility(IEncrypterClass) is self.__class__:
                _gsm.registerUtility(self)
            _gsm.registerUtility(self, name=self.name)
        except ComponentLookupError:
            pass

    @classmethod
    def register_class(self, default=False):
        _gsm.registerUtility(self, IEncrypterClass, name=self.name)
        if default:
            _gsm.registerUtility(self, IEncrypterClass)


def encrypt(data, no_cipher_name=False):
    try:
        return _gsm.getUtility(IEncrypter).encrypt(data, no_cipher_name=no_cipher_name)
    except ComponentLookupError:
        return data


def get_encryption_signature(data):
    if data.startswith(b".e"):
        return data[2:data.find(b"$")]
    else:
        return None


def decrypt(data):
    sig = get_encryption_signature(data)
    if sig is not None:
        # Named utility if encrypter name is known, default if ''
        return _gsm.getUtility(IEncrypter, sig.decode()).decrypt(data)
    else:
        return data


def init(**kw):
    for name, cls in _gsm.getUtilitiesFor(IEncrypterClass):
        if name:
            try:
                utility = _gsm.getUtility(IEncrypter, name=name)
                _gsm.unregisterUtility(utility)
            except ComponentLookupError:
                pass
            cls(**kw).register()



"""
Pytest fixtures for zerodb tests

To activate the fixtures, add this to your conftest.py:

from zerodb.testing import *
"""
import os
import pytest
import shutil
import tempfile

import ZEO.tests.testssl

import zerodb
from zerodb.crypto import kdf

TEST_PASSPHRASE = "v3ry 53cr3t pa$$w0rd"


__all__ = [
    "TEST_PASSPHRASE",
    "tempdir",
    "do_zeo_server",
    "db",
]


@pytest.fixture(scope="module")
def tempdir(request):
    tmpdir = tempfile.mkdtemp()
    request.addfinalizer(lambda: shutil.rmtree(tmpdir))
    return tmpdir


@pytest.fixture(scope="module")
def db(request, zeo_server, dbclass=zerodb.DB):
    zdb = dbclass(zeo_server,
                  username='root', password=TEST_PASSPHRASE,
                  security=kdf.key_from_password,
                  server_cert=ZEO.tests.testssl.server_cert,
                  debug=True, wait_timeout=11)

    if request is not None:
        @request.addfinalizer
        def fin():
            zdb.disconnect()  # I suppose, it's not really required

    return zdb


def do_zeo_server(request, tempdir, name=None, fsname='db.fs'):
    sock, stop = zerodb.server(
            name=name, path=os.path.join(tempdir, fsname),
            init=dict(
                password=TEST_PASSPHRASE, cert=ZEO.tests.testssl.client_cert
                ))
    request.addfinalizer(stop)
    return sock



from . import compress_common
from . import encrypt_common

compress = compress_common.compress
decompress = compress_common.decompress
encrypt = encrypt_common.encrypt
decrypt = encrypt_common.decrypt
init_crypto = encrypt_common.init
get_encryption_signature = encrypt_common.get_encryption_signature



import persistent
from zope.index.text import widcode


class PersistentWid(persistent.Persistent):
    """
    Behaves like a string, but stored as a lazy-loaded persistent object.
    Can be encoded from word ids
    """

    def __init__(self, s):
        self.s = s

    @classmethod
    def encode_wid(cls, l):
        return cls(widcode.encode(l))

    def decode_wid(self):
        return widcode.decode(self.s)

    def __getattribute__(self, attr):
        try:
            return super(PersistentWid, self).__getattribute__(attr)
        except AttributeError:
            return super(PersistentWid, self).__getattribute__('s').__getattribute__(attr)



from zope.interface import implementer
from zerodb.collective.indexing.interfaces import IIndexQueueProcessor


class IPortalCatalogQueueProcessor(IIndexQueueProcessor):
    """ an index queue processor for the standard portal catalog via
        the `CatalogMultiplex` and `CMFCatalogAware` mixin classes """


@implementer(IPortalCatalogQueueProcessor)
class PortalCatalogProcessor(object):

    def index(self, obj, attributes=None):
        #index(obj, attributes)
        pass

    def reindex(self, obj, attributes=None):
        #reindex(obj, attributes)
        pass

    def unindex(self, obj):
        #unindex(obj)
        pass

    def begin(self):
        pass

    def commit(self):
        pass

    def abort(self):
        pass



#!/usr/bin/env python

# Warning!
# At the moment, this API is safe to use only on same machine
# (no https etc)

import imp
import jsonpickle
import six
import transaction
import zerodb
from flask import Flask, jsonify, session, request
from flask.ext.cors import CORS
from zerodb.catalog.query import optimize
from zerodb.catalog import query_json as qj

try:
    import simplejson as json
except:
    import json

PORT = 2015
HOST = "127.0.0.1"
DEBUG = True
DEV_SECRET_KEY = "development key"

app = Flask(__name__)
cors = CORS(app, supports_credentials=True)
dbs = {}
models = None
zeosocket = None


@app.route("/_connect", methods=["GET", "POST"])
def connnect():
    if request.method == "GET":
        req = request.args
    elif request.method == "POST":
        req = request.form
    else:
        return jsonify(ok=0)

    username = req.get("username")
    passphrase = req.get("passphrase")
    if zeosocket:
        socket = zeosocket
    else:
        host = req.get("host")
        port = req.get("port")
        if host and port:
            socket = (host, port)
        elif host:
            socket = host
        else:
            socket = None

    if not (username and passphrase and socket):
        return jsonify(ok=0, message="Incomplete login information")

    try:
        db = zerodb.DB(socket, username=username, password=passphrase)
    except Exception as e:
        return jsonify(ok=0, message=str(e), error_type=e.__class__.__name__)

    session["username"] = username
    dbs[username] = db

    return jsonify(ok=1)


@app.route("/<table_name>/_find", methods=["GET", "POST"])
def find(table_name):
    db = dbs[session["username"]]
    model = getattr(models, table_name)

    if request.method == "GET":
        req = request.args
    elif request.method == "POST":
        req = request.form
    else:
        return jsonify(ok=0)

    if req.get("criteria") is None:
        criteria = json.loads(request.data).get("criteria")
    else:
        criteria = json.loads(req.get("criteria"))

    if isinstance(criteria, dict) and (len(criteria) == 1) and "_id" in criteria:
        ids = [c["$oid"] for c in criteria["_id"]]
    else:
        ids = None
        criteria = optimize(qj.compile(criteria))

    skip = req.get("skip")
    if skip:
        skip = int(skip)

    limit = req.get("limit")
    if limit:
        limit = int(limit)

    sort = req.get("sort")
    if sort:
        try:
            sort = json.loads(sort)
        except ValueError:
            if sort.startswith("-"):
                sort_index = sort[1:].strip()
                reverse = True
            else:
                sort_index = sort
                reverse = None
        if isinstance(sort, dict):
            assert len(sort) == 1  # Only one field at the moment
            sort_index, direction = sort.popitem()
            reverse = (direction >= 0)
        elif isinstance(sort, list):
            sort_index = sort[0]
            reverse = None
    else:
        sort_index = None
        reverse = None

    if ids:
        skip = skip or 0
        end = skip + limit if limit else None
        ids = ids[skip:end]
        result = db[model][ids]
    else:
        result = db[model].query(criteria, skip=skip, limit=limit, sort_index=sort_index, reverse=reverse)

    return jsonpickle.encode(list(result), unpicklable=False)


@app.route("/<table_name>/_remove", methods=["GET", "POST"])
def remove(table_name):
    db = dbs[session["username"]]
    model = getattr(models, table_name)

    if request.method == "GET":
        req = request.args
    elif request.method == "POST":
        req = request.form
    else:
        return jsonify(ok=0)

    criteria = req.get("criteria")
    ids = req.get("_id")
    if criteria:
        criteria = json.loads(req.get("criteria"))
        if isinstance(criteria, dict) and (len(criteria) == 1) and "_id" in criteria:
            ids = [c["$oid"] for c in criteria["_id"]]
        else:
            ids = None
            criteria = optimize(qj.compile(criteria))
        result = db[model].query(criteria)
    elif ids:
        ids = json.loads(ids)
        result = db[model][ids]
    else:
        return jsonify(ok=0)

    try:
        with transaction.manager:
            count = db.remove(result)
    except Exception as e:
        return jsonify(ok=0, message=str(e), error_type=e.__class__.__name__)

    return jsonify(ok=1, count=count)


@app.route("/<table_name>/_insert", methods=["GET", "POST"])
def insert(table_name):
    # POST has array of documents or one document
    db = dbs[session["username"]]
    model = getattr(models, table_name)

    if request.method == "POST":
        try:
            if request.form.get("docs") is None:
                data = json.loads(request.data).get("docs")
            else:
                data = json.loads(request.form.get("docs"))
            if isinstance(data, dict):
                data = [data]
            objs = [model(**row) for row in data]
            with transaction.manager:
                oids = [{"$oid": db.add(o)} for o in objs]
            return jsonify(status={"ok": 1}, oids=oids)
        except Exception as e:
            return jsonify(ok=0, message=str(e), error_type=e.__class__.__name__)
    else:
        return jsonify(ok=0)


@app.route("/<table_name>/_get", methods=["GET", "POST"])
def get(table_name):
    db = dbs[session["username"]]
    model = getattr(models, table_name)

    if request.method == "GET":
        req = request.args
    elif request.method == "POST":
        req = request.form
    else:
        return jsonify(ok=0)

    data = json.loads(req["_id"])

    return jsonpickle.encode(db[model][data], unpicklable=False)


@app.route("/_disconnect", methods=["GET", "POST"])
def disconnect():
    if "username" in session:
        username = session.pop("username")
        if username in dbs:
            del dbs[username]
    return jsonify(ok=1)


def run(data_models=None, host=HOST, port=PORT, debug=DEBUG, secret_key=DEV_SECRET_KEY, zeo_socket=None, **kw):
    global models
    global zeosocket

    if isinstance(data_models, six.string_types):
        models = imp.load_source("models", data_models)
    else:
        models = data_models

    zeosocket = zeo_socket

    app.config["SECRET_KEY"] = secret_key
    app.run(host=host, port=port, debug=debug, **kw)


if __name__ == "__main__":
    run()



# We don't need two-way references given by IntIds, but we need a primary object storage
# If we used IntIds and some objects storage we'd have *3* trees
# But we want to minimize number of calls, so we make one primary IOBTree which has IDs and objects
# Conflict avoidance method is copied from IntIds._generateId (need something better I guess)
import persistent
import random

import six
from BTrees.Length import Length
from zerodb.trees import family32


class IdStore(persistent.Persistent):
    """
    Persistent storage of objects which generates non-conflictiong unique IDs
    for them
    """

    _v_nextid = None

    family = family32

    def __init__(self, family=family32):
        """
        :param family: Family of BTrees to use
        """
        self.tree = family.IO.BTree()
        self.length = Length()

    def _generateId(self):
        """Generate an id which is not yet taken.

        This tries to allocate sequential ids so they fall into the same BTree
        bucket, and randomizes if it stumbles upon a used one.

        This algorithm is taken from zope.intid but it will cause performance
        degradation due to fragmentation if used too often, we need something
        better eventually
        """
        nextid = self._v_nextid
        while True:
            if nextid is None:
                nextid = random.randrange(0, self.family.maxint)
            uid = nextid
            if uid not in self.tree:
                nextid += 1
                if nextid > self.family.maxint:
                    nextid = None
                self._v_nextid = nextid
                return uid
            nextid = None

    def add(self, obj):
        """
        Add object to the storage

        :param obj: Object to store (persistent.Persistent but not necessarily)
        :return: Unique ID
        :rtype: int
        """
        if not hasattr(self, "length"):
            self.length = Length(len(self.tree))
        while True:
            uid = self._generateId()
            if self.tree.insert(uid, obj):  # We use this feature of BTrees to avoid conflicts
                # _v_* and _p_* are not saved in the database
                # However assigning _v_* unghostifies the object while _p_* doesn't
                # We don't want to force-unghostify the object in some cases
                obj._p_uid = uid
                self.length.change(1)
                return uid

    def remove(self, iobj):
        """
        Remove object from the storage
        :param obj: Object or its integer unique id
        """
        if not hasattr(self, "length"):
            self.length = Length(len(self.tree))
        if type(iobj) in six.integer_types:
            del self.tree[iobj]
            self.length.change(-1)
        elif hasattr(iobj, "_p_uid"):
            del self.tree[iobj._p_uid]
            iobj._p_uid = None
            self.length.change(-1)
        else:
            raise TypeError("Argument should be either uid or object itself")

    def __getitem__(self, uid):
        """
        :param int uid: Get object by its unique ID
        """
        return self.tree[uid]

    def __delitem__(self, uid):
        """
        :param int uid: Get object by its unique ID
        """
        self.remove(uid)

    def __len__(self):
        if not hasattr(self, "length"):
            self.length = Length(len(self.tree))
        return self.length.value



"""User database management

The concept of"root" is a little tricky, so we try to avoid using it
for admin purposes.  For this reason, we add an Admin object ro the
root and assure that it has oid 1, allowing us to navigate to it
without going through the root.

Database root objects::

  users: {userid -> User}
  users_by_der: {der -> User}
  certs: Certs

Where DER is a DER encoding of a cert.

The root user's user id is z64 (aka 0).

All other user's ids are the oids of their root folders.

Users have certs: {der -> pem_data}

The Certs object is just a persistent container for the concatenation
of all of the user certs.

"""
import hashlib
import os
import ssl
import uuid

from BTrees.OOBTree import BTree
from ZODB.utils import p64
import persistent
import persistent.mapping
import ZODB
import ZODB.FileStorage

from zerodb.crypto import kdf

from .ownerstorage import OwnerStorage

ONE = p64(1)


def get_der(pem_data):
    context = ssl.create_default_context(cadata=pem_data)
    [cert_der] = context.get_ca_certs(1)  # TCBOO
    return cert_der


def hash_password(password, salt):
    if not isinstance(password, bytes):
        password = password.encode()
    if not isinstance(salt, bytes):
        salt = salt.encode()
    return b'sha256::' + hashlib.sha256(password + salt).digest()


class User(persistent.Persistent):

    password = None

    def __init__(self, name, root, password=None):
        """
        :param str od: User id
        :param str name: User name
        :param PersistentMapping: User's database root
        """
        self.name = name
        self.root = root
        self.id = root._p_oid
        # Today, TCBOO cert, but maybe later
        self.certs = {}  # {cert_der -> cert_pem}

        if password:
            self.salt = uuid.uuid4().hex.encode()
            self.password = hash_password(password, self.salt)

    def check_password(self, password):
        return hash_password(password, self.salt) == self.password

    def change_password(self, password):
        if password is not None:
            if password:
                self.password = hash_password(password, self.salt)
            else:
                self.password = None


class Certs(persistent.Persistent):

    def __init__(self):
        self.data = ''

    def add(self, pem_data):
        self.data += '\n\n' + pem_data

    def remove(self, pem_data):
        self.data = self.data.replace('\n\n' + pem_data, '')


class Admin(persistent.Persistent):

    def __init__(self, conn):
        conn.add(self)
        assert self._p_oid == ONE

        self.users         = BTree()  # {uid -> user}
        self.users_by_name = BTree()  # {uname -> user}
        self.uids          = BTree()  # {cert_der -> uid}
        self.certs         = Certs()  # Cert, persistent wrapper for
                                      # concatinated cert data

        # Add nobody placeholder
        with open(os.path.join(os.path.dirname(__file__), 'nobody.pem')) as f:
            nobody_pem = f.read()

        self.certs.add(nobody_pem)
        self.uids[get_der(nobody_pem)] = None

    def add_user(self, uname, pem_data=None, password=None,
                 security=kdf.hash_password, appname='zerodb.com'):
        root = persistent.mapping.PersistentMapping()
        self._p_jar.add(root)

        password, _ = security(
                uname, password,
                key_file=None, cert_file=None,
                appname=appname, key=None)

        user = User(uname, root, password)
        self.users[user.id] = user
        self.users_by_name[user.name] = user

        if pem_data:
            self._add_user_cert(user, pem_data)

        return user

    def _add_user_cert(self, user, pem_data):
        cert_der = get_der(pem_data)
        if cert_der in self.uids or cert_der in user.certs:
            raise ValueError("SSL certificate id already used",
                             pem_data, user.id)
        self.uids[cert_der] = user.id
        user.certs[cert_der] = pem_data
        self.certs.add(pem_data)

    def _del_user_certs(self, user):
        for der, pem_data in user.certs.items():
            del self.uids[der]
            self.certs.remove(pem_data)

    def del_user(self, name):
        user = self.users_by_name.pop(name)
        del self.users[user.id]
        self._del_user_certs(user)

    def change_cert(self, name, pem_data=None, password=None,
                    security=kdf.hash_password, appname='zerodb.com'):
        user = self.users_by_name[name]

        if pem_data is not None:
            self._del_user_certs(user)
            user.certs.clear()
            if pem_data:
                self._add_user_cert(user, pem_data)

        if password is not None:
            password, _ = security(
                    name, password,
                    key_file=None, cert_file=None,
                    appname=appname, key=None)
            user.change_password(password)


def get_admin(conn):
    return conn.get(ONE)


def init_db(storage, uname, pem_data=None, close=True, password=None):
    db = ZODB.DB(OwnerStorage(storage, p64(2)))
    with db.transaction() as conn:
        conn.root.admin = Admin(conn)
        user = conn.root.admin.add_user(uname, pem_data, password)
        assert user.id == db.storage.user_id
    if close:
        db.close()


def init_db_script():
    import argparse
    import os

    parser = argparse.ArgumentParser(
        description="Create an initialized ZeroDB file-storage with a root user"
        )
    parser.add_argument("path", help="Path for new file-storage file")
    parser.add_argument("user", help="Name of root user")
    parser.add_argument("certificate", help="Path to user certificate")

    options = parser.parse_args()

    path = options.path
    if os.path.exists(path):
        raise ValueError("Path exists", path)

    with open(options.certificate) as f:
        pem_data = f.read()

    init_db(ZODB.FileStorage.FileStorage(path), options.user, pem_data)



from .db import DB

def server(path=None, blob_dir=None, storage_conf=None, zeo_conf=None,
           port=0, threaded=True, **kw):
    import os, zerodb.forker
    if storage_conf is None and path is None:
        storage_conf = '<mappingstorage>\n</mappingstorage>'

    return zerodb.forker.start_server(
        storage_conf, zeo_conf, port, keep=True, path=path,
        blob_dir=blob_dir, suicide=False, threaded=threaded, **kw)



from ZODB.POSException import POSKeyError, StorageError
from ZODB.utils import maxtid, u64, z64
import ZODB.interfaces
import zope.interface

@zope.interface.implementer(ZODB.interfaces.IMultiCommitStorage)
class OwnerStorage(object):
    """Storage wrapper that adds/stript/checks owner id in record

    Some special considerations around bootstrapping:

    - Except for the root users, users' ids are their root folder's
      oids.  This allows us to grant access even when a user's root
      folder was written by the root user.  Note that once a user
      modified their root folder, the root user won't have access.

    - All user can access object 0, but they can't access any
      persistent subobjects of it.
    """

    # methods we punt on:
    history = load = restore = iterator = undo = undoLog = undoInfo = None
    record_iternext = storeBlob = loadBlob = restoreBlob = None
    def __iter__(self):
        if False: yield
    def supportsUndo(self):
        return False

    def __init__(self, storage, user_id):
        self.user_id = user_id
        self.storage = storage

    def __getattr__(self, name):
        return getattr(self.storage, name)

    def _check_permissions(self, data, oid=None):
        if not (
            data.endswith(self.user_id) or
            oid == self.user_id or
            oid == z64
            ):
            raise StorageError(
                "Attempt to access encrypted data of others at <%s> by <%s>" % (
                    u64(oid), u64(self.user_id)))

    def loadBefore(self, oid, tid):
        r = self.storage.loadBefore(oid, tid)
        if r is not None:
            data, serial, after = r
            self._check_permissions(data, oid)
            return data[:-len(self.user_id)], serial, after
        else:
            return r

    def loadSerial(self, oid, serial):
        r = self.storage.loadSerial(oid, tid)
        data = self.storage.loadSerial(self, oid, serial)
        self._check_permissions(data, oid)
        return data[:-len(self.user_id)]

    def store(self, oid, serial, data, version, transaction):
        try:
            old_data = self.storage.loadBefore(oid, maxtid)[0]
            self._check_permissions(old_data, oid)
        except POSKeyError:
            pass  # We store a new one
        data += self.user_id
        self.storage.store(oid, serial, data, version, transaction)

    def __len__(self):
        return len(self.storage)



from logging import getLogger
from threading import local
from zope.interface import implementer
from zope.component import getSiteManager

from zerodb.collective.indexing.interfaces import IIndexQueue
from zerodb.collective.indexing.interfaces import IIndexQueueProcessor
from zerodb.collective.indexing.config import INDEX, REINDEX, UNINDEX
from zerodb.collective.indexing.transactions import QueueTM

debug = getLogger('collective.indexing.queue').debug


localQueue = None
processing = set()


class InvalidQueueOperation(Exception):
    pass


def getQueue():
    """ return a (thread-local) queue object, create one if necessary """
    global localQueue
    if localQueue is None:
        localQueue = IndexQueue()
    return localQueue


def processQueue():
    """ process the queue (for this thread) immediately """
    queue = getQueue()
    processed = 0
    if queue.length() and not queue in processing:
        debug('auto-flushing %d items: %r', queue.length(), queue.getState())
        try:
            processing.add(queue)
            processed = queue.process()
        finally:
            processing.remove(queue)
    return processed


@implementer(IIndexQueue)
class IndexQueue(local):

    def __init__(self):
        self.queue = []
        self.tmhook = None

    def hook(self):
        """ register a hook into the transaction machinery if that hasn't
            already been done;  this is to make sure the queue's processing
            method gets called back just before the transaction is about to
            be committed """
        if self.tmhook is None:
            self.tmhook = QueueTM(self).register
        self.tmhook()

    def index(self, obj, attributes=None):
        self.queue.append((INDEX, obj, attributes))
        self.hook()

    def reindex(self, obj, attributes=None):
        self.queue.append((REINDEX, obj, attributes))
        self.hook()

    def unindex(self, obj):
        self.queue.append((UNINDEX, obj, None))
        self.hook()

    def setHook(self, hook):
        self.tmhook = hook

    def getState(self):
        return list(self.queue)     # better return a copy... :)

    def setState(self, state):
        self.queue = state

    def length(self):
        """ return number of currently queued items;  please note that
            we cannot use `__len__` here as this will cause test failures
            due to the way objects are compared """
        return len(self.queue)

    def optimize(self):
        res = {}
        for iop, obj, iattr in self.getState():
            oid = hash(obj)
            func = getattr(obj, 'getPhysicalPath', None)
            if callable(func):
                oid = oid, func()
            op, dummy, attr = res.get(oid, (0, obj, iattr))
            # If we are going to delete an item that was added in this transaction, ignore it
            if op == INDEX and iop == UNINDEX:
                del res[oid]
            else:
                # Operators are -1, 0 or 1 which makes it safe to add them
                op += iop
                op = min(max(op, UNINDEX), INDEX)  # operator always between -1 and 1

                # Handle attributes, None means all fields, and takes presedence
                if isinstance(attr, (tuple, list)) and isinstance(iattr, (tuple, list)):
                    attr = tuple(set(attr).union(iattr))
                else:
                    attr = None

                res[oid] = (op, obj, attr)

        debug('finished reducing; %d item(s) in queue...', len(res))
        # Sort so unindex operations come first
        self.setState(sorted(res.values()))

    def process(self):
        self.optimize()
        if not self.queue:
            return 0
        sm = getSiteManager()
        utilities = list(sm.getUtilitiesFor(IIndexQueueProcessor))
        processed = 0
        for name, util in utilities:
            util.begin()
        # TODO: must the queue be handled independently for each processor?
        for op, obj, attributes in self.queue:
            for name, util in utilities:
                if op == INDEX:
                    util.index(obj, attributes)
                elif op == REINDEX:
                    util.reindex(obj, attributes)
                elif op == UNINDEX:
                    util.unindex(obj)
                else:
                    raise InvalidQueueOperation(op)
            processed += 1
        debug('finished processing %d items...', processed)
        self.clear()
        return processed

    def commit(self):
        sm = getSiteManager()
        for name, util in sm.getUtilitiesFor(IIndexQueueProcessor):
            util.commit()

    def abort(self):
        sm = getSiteManager()
        for name, util in sm.getUtilitiesFor(IIndexQueueProcessor):
            util.abort()
        self.clear()

    def clear(self):
        del self.queue[:]
        # release transaction manager
        self.tmhook = None



from zope.interface import Interface


class IIndexing(Interface):
    """ interface for indexing operations, used both for the queue and
        the processors, which perform the actual indexing;  the queue gets
        registered as a utility while the processors (portal catalog, solr)
        are registered as named utilties """

    def index(obj, attributes=None):
        """ queue an index operation for the given object and attributes """

    def reindex(obj, attributes=None):
        """ queue a reindex operation for the given object and attributes """

    def unindex(obj):
        """ queue an unindex operation for the given object """


class IIndexQueue(IIndexing):
    """ a queue for storing and optimizing indexing operations """

    def setHook(hook):
        """ set the hook for the transaction manager;  this hook must be
            called whenever an indexing operation is added to the queue """

    def getState():
        """ get the state of the queue, i.e. its contents """

    def setState(state):
        """ set the state of the queue, i.e. its contents """

    def process():
        """ process the contents of the queue, i.e. start indexing;
            returns the number of processed queue items """

    def clear():
        """ clear the queue's contents in an ordered fashion """


class IIndexQueueProcessor(IIndexing):
    """ a queue processor, i.e. an actual implementation of index operations
        for a particular search engine, e.g. the catalog, solr etc """

    def begin():
        """ called before processing of the queue is started """

    def commit():
        """ called after processing of the queue has ended """

    def abort():
        """ called if processing of the queue needs to be aborted """



import BTrees
import itertools
import logging

import six
from BTrees.Length import Length
from BTrees.OOBTree import TreeSet as SortedSet  # sorted set implemented in C
from collections import Counter, defaultdict
from itertools import islice
from six.moves import zip as izip, map as imap, xrange
from math import sqrt, log
from persistent import Persistent
from zope.interface import implementer
from zope.index.interfaces import IInjection
from zope.index.interfaces import IStatistics
from zope.index.text.interfaces import IExtendedQuerying
from zope.index.text.interfaces import ILexiconBasedIndex
from zope.index.text.queryparser import _tokenizer_regex

from zerodb.storage import prefetch, parallel_traversal
from zerodb.catalog.indexes.pwid import PersistentWid


class LengthyTree(object):
    def __init__(self, obj, L):
        """
        Replaces __len__ method to use Length, proxies all other methods.

        :param Persistent obj: tree-like object
        :param Length L: BTrees' Length object
        """
        self.__obj = obj
        self.__L = L

    def __len__(self):
        return self.__L.value

    def __getattr__(self, attr):
        return getattr(self.__obj, attr)

    def __iter__(self):
        return self.__obj.__iter__()


@implementer(IInjection, IStatistics, ILexiconBasedIndex, IExtendedQuerying)
class IncrementalLuceneIndex(Persistent):
    """
    Using Lucene's practical scoring function and incremental search
    """
    # Using Okapi index requires us to fetch all docids for all word ids.
    # That makes scalability of the search bad, even for a single keyword
    # This class uses Lucene's practical scoring function:
    # https://www.elastic.co/guide/en/elasticsearch/guide/current/practical-scoring-function.html

    # In brief, practical scoring is this:
    #   t - keyword (or term) which is searched
    #   D - document
    #   TF(t, D) = sqrt(f(t, D))
    #   IDF(t) = 1 + log(N_docs / (N_docs_for_term + 1))
    #   score(term, D) = TF * IDF**2 / sqrt(N_terms_in_D)
    #   score(D) = sum(score(term, D)) / sqrt(sum(IDF(t)**2))

    # When we search for one keyword, we can keep document UIDs
    # Ordered by w_t = TF(t, D) / sqrt(N_terms_in_D),
    # for example in a TreeSet((-w_t, uid)).
    # Reading it incrementally gives docs ordered from most to least relevant.

    # Now the harder part (not in Lucene)

    # When we want to do a multi-keyword search, it is, in a way, weighted union
    # of searches for all keywords.
    # E.g. now:
    #   w(D) = sum(w_t * IDF(t)**2)   // no need to normalize by sqrt(sum(idf**2))
    # weight calculation and sort in memory, which fetches docweights for all terms again.
    # This is slow when e2e encrypted and fetched remotely, and also reveals sum(N_docs_for_term) to the server

    # Here's how we do the same incrementally ("lazy")
    # We allow elements to be shifted by max_out_of_order = 3 from their ideal position.
    # We read beginnings of sorted lists of (w_i, uid_i) for each word.
    # After we calculate the sum,
    # w_j_known = sum(w_j * IDF(t_j)**2, t: all terms for which we know w_j for uid_i)

    # for i_th element the weight could be between:
    #   w_min_i(t_j) = sum(w_min(t_j) * IDF(t_j)**2: all terms for which we don't know w_j) + w_i_known
    #   and
    #   w_max_i(t_j) = sum(w_max(t_j) * IDF(t_j)**2: all terms for which we don't know w_j) + w_i_known

    # For each term, we have current window of docs ordered by weight which still haven't been returned.
    # It is likely that some term is particularly selective and largely outweights others,
    # E.g. for almost all elements in the current window, weight ranges (w_min_i, w_max_i) don't overlap
    # for different i. Or, perhaps, elements are misplaced by no more than max_out_of_order positions.

    # We choose a term for which all positions at the tail of the window can never appear within the limit # of docs.
    # If there are none, we expand most promising wid set by factor of 2 until we get this situation.
    # Out of those terms (if there are multiple), we select one which has smallest number of
    # doc uids misplaced.

    # For misplaced docids, we fetch their weights. Now we know order within the specified
    # number of documents.

    # For this to work, we need to keep
    # BTree(wids -> (TreeSet(w_i, uid_i), BTree(uid_i -> w_i), Length))
    # Length is for calculating IDF

    family = BTrees.family32
    lexicon = property(lambda self: self._lexicon,)
    # * index_doc
    # * unindex_doc
    # * clear
    # * documentCount
    # * wordCount
    # * lexicon
    # search
    # search_phrase
    # search_glob
    # _search_wids
    # * query_weight

    def __init__(self, lexicon, family=None, keep_phrases=True):
        if family is not None:
            self.family = family
        self._lexicon = lexicon
        self.keep_phrases = keep_phrases
        self.clear()

    def clear(self):
        # ._wordinfo = BTree(wids -> (TreeSet((weight, docid)), BTree(docid -> weight), Length))
        self._wordinfo = self.family.IO.BTree()

        # ._docwords = BTree(docid -> widcode)
        # used for document unindexing
        # but no phrase search
        self._docwords = self.family.IO.BTree()

        self.wordCount = Length()
        self.documentCount = Length()

    def _get_doctrees(self, wids):
        """
        Gets persistent objects used for indexes for wids
        returns: {wid -> TreeSet((weight, docid))}, {wid -> Length}
        """
        weights = {}
        lengths = {}
        parallel_traversal(self._wordinfo, wids)

        for wid in wids:
            record = self._wordinfo.get(wid)
            if record is None:
                length = Length(0)
                wdocid = self.family.OO.TreeSet()
                self._wordinfo[wid] = (wdocid, length)
                self.wordCount.change(1)
            else:
                wdocid, length = record

            weights[wid] = wdocid
            lengths[wid] = length

        return weights, lengths

    def get_words(self, docid):
        """Return a list of the wordids for a given docid."""
        return self._docwords[docid].decode_wid()

    def _get_widscores(self, ctr, docid, widlen=None):
        """
        Get scores per word:
            score = sqrt(f) / sqrt(widlen)

        :param dict ctr: Counter {wid -> number_of_wids}
        :param int widlen: Number of unique words in document. Use len(ctr) if
            not given
        :returns dict: {wid -> (-score, docid)}
        """
        widlen = sqrt(widlen or len(ctr))
        return {w: (-sqrt(f) / widlen, docid) for w, f in ctr.items()}

    def index_doc(self, docid, text):
        if docid in self._docwords:
            return self._reindex_doc(docid, text)

        wids = self._lexicon.sourceToWordIds(text)

        # XXX Counter is slow. If it is an issue, need to include C module
        # http://stackoverflow.com/questions/2522152/python-is-a-dictionary-slow-to-find-frequency-of-each-character
        widcnt = Counter(wids)
        widset = widcnt.keys()
        widcode = PersistentWid.encode_wid(wids if self.keep_phrases else widset)
        self._docwords[docid] = widcode

        if widset:
            weights, lengths = self._get_doctrees(widset)
            docscores = self._get_widscores(widcnt, docid)
            parallel_traversal(*zip(*[(weights[w], docscores[w]) for w in widset]))
            prefetch(list(lengths.values()) + [self.documentCount])

            for w in widset:
                weights[w].add(docscores[w])
                lengths[w].change(1)

        self.documentCount.change(1)

        return len(wids)

    def _reindex_doc(self, docid, text):
        # We should change Length only for new wids used
        old_wids = self.get_words(docid)
        old_ctr = Counter(old_wids)
        old_widset = set(old_ctr)
        new_wids = self._lexicon.sourceToWordIds(text)
        new_ctr = Counter(new_wids)
        new_widset = set(new_ctr)
        removed_wids = old_widset - new_widset
        added_wids = new_widset - old_widset
        all_wids = list(new_widset | old_widset)

        weights, lengths = self._get_doctrees(all_wids)

        for w in removed_wids:
            lengths[w].change(-1)
        for w in added_wids:
            lengths[w].change(1)

        old_docscores = self._get_widscores(old_ctr, docid)
        new_docscores = self._get_widscores(new_ctr, docid)
        parallel_traversal(*zip(*[
            (weights[w], old_docscores.get(w) or new_docscores.get(w))
            for w in all_wids]))
        # We should update all the weights if len(old_wids) != len(new_wids)
        # ...and that is generally the case, so we update always
        for w in old_widset:
            try:
                weights[w].remove(old_docscores[w])
            except KeyError:
                # This should never happen
                # If it does, it's a bad sign, though we should still work
                logging.error("Old weight-docid pair not found!")
        for w in new_widset:
            weights[w].add(new_docscores[w])

        self._docwords[docid] = PersistentWid.encode_wid(new_wids if self.keep_phrases else new_widset)

        return len(new_wids)

    def unindex_doc(self, docid):
        if docid not in self._docwords:
            return
        wids = self.get_words(docid)
        ctr = Counter(wids)
        wids = list(ctr)
        weights, lengths = self._get_doctrees(wids)
        scores = self._get_widscores(ctr, docid)
        parallel_traversal(*zip(*[(weights[w], scores[w]) for w in wids]))
        for w in wids:
            lengths[w].change(-1)
            weights[w].remove(scores[w])
            if lengths[w].value == 0:
                del self._wordinfo[w]
        del self._docwords[docid]
        self.documentCount.change(-1)

    def idf2(self, wid):
        """
        Returns IDF squared.
        Note that definition of IDF is different from the classic one
        """
        N_docs = self.documentCount.value
        N_for_term = self._wordinfo[wid][1].value
        return (1.0 + log(N_docs / (1.0 + N_for_term))) ** 2

    def _remove_oov_wids(self, wids):
        parallel_traversal(self._wordinfo, set(wids))
        return filter(self._wordinfo.has_key, wids)

    def query_weight(self, terms):
        """
        terms - string with terms or list with wids

        Normalization factor:
            sum of idfs squared (with number-of-uses coefficient)
        """
        if isinstance(terms, list):
            wc = Counter(terms)
        else:
            wc = Counter(self._lexicon.termToWordIds(terms))
        if 0 in wc:
            del wc[0]
        wids = self._remove_oov_wids(wc.keys())
        return sum([self.idf2(w) * wc[w] for w in wids])

    def _search_wids(self, wids):
        """
        Finds pointers to iterables (-score, docid) for wids in order,
        IDFs squared are also calculated as weights
        """
        return [(LengthyTree(*self._wordinfo[w]), self.idf2(w)) for w in wids]

    def search(self, term):
        wids = self._lexicon.termToWordIds(term)
        wids = self._remove_oov_wids(wids)
        if not wids:
            return []
        return mass_weightedUnion(self._search_wids(wids))

    def search_glob(self, pattern):
        wids = self._lexicon.globToWordIds(pattern.lower())
        wids = self._remove_oov_wids(wids)
        return mass_weightedUnion(self._search_wids(wids))

    def _search_all(self, term):
        tokens = [t.lower() for t in _tokenizer_regex.findall(term)]
        glob_cond = lambda t: ('?' in t) or ('*' in t)
        glob_tokens = filter(glob_cond, tokens)
        word_tokens = filter(lambda t: not glob_cond(t), tokens)
        wids = set()
        wids.update(self._lexicon.termToWordIds(word_tokens))
        wids.update(itertools.chain(*map(self._lexicon.globToWordIds, glob_tokens)))
        wids = self._remove_oov_wids(wids)
        # XXX
        # We should have OrderedDict-like lazy objects
        # and we should have weightedIntersection and weightedUnion
        # working lazily in a for of zerodbext.catalog
        # This is just a workaround for simpler queries
        # XXX
        return imap(lambda x: x[0], mass_weightedUnion(self._search_wids(wids)))

    def search_phrase(self, phrase):
        # Need to do mass_weightedIntersection here.
        # Simplest version of it would be:
        #   fetch (score, docid) info for the rarest keyword
        #   find same docids in different keywords
        #   do intersection in memory
        #   fetch widcode to search for phrase
        # XXX we'd require to maintain also BTree(docid -> score)
        raise NotImplementedError


def mass_weightedUnion(L):
    """
    Incremental version of mass_weightedUnion
    :param list L: (TreeSet((-score, docid)), weight) elements
    :returns: iterable ordered from large to small sum(score*weight)
    """
    cache_size = 40
    order_size = 15
    order_violation = 3
    cache_updated = None

    if len(L) == 0:
        return

    elif len(L) == 1:
        # Trivial
        tree, weight = L[0]
        # XXX need to make it possible to advance the tree N elements ahead!
        for (score, docid) in tree:
            yield (docid, -score * weight)

    else:
        # XXX make into an iterator class

        trees, weights = zip(*L)
        prefetch(trees)
        prefetch([x._firstbucket for x in trees if x._firstbucket is not None])

        unread_max = [-t.minKey()[0] * w for t, w in L]
        lengths = map(len, trees)
        iters = dict(enumerate(map(iter, trees)))
        caches = [{} for i in range(len(L))]
        docid2cacheid = defaultdict(list)
        cache_len = None
        maxscores = [-1] * len(L)
        used = set()
        sorted_mins = SortedSet()  # Contains tuples (-min_weight, docid)
        mins_dict = {}  # {docid -> min_weight}
        docids = []

        def precache(i, size):
            try:
                for j in xrange(size):
                    score, docid = next(iters[i])
                    score = -score * weights[i]
                    if unread_max[i] > score:
                        unread_max[i] = score
                    if docid not in used:
                        caches[i][docid] = score
                        docid2cacheid[docid].append(i)
                        if maxscores[i] < 0:
                            maxscores[i] = score

                        # documents are not repeated in iterators
                        # it means that for the iterator i we didn't meet docid before
                        # which means that we should just add score to the current minscore
                        total_score = mins_dict.get(docid, 0.0) + score
                        if docid in mins_dict:
                            sorted_mins.remove((-mins_dict[docid], docid))
                        mins_dict[docid] = total_score
                        sorted_mins.add((-total_score, docid))
            except StopIteration:
                del iters[i]
                unread_max[i] = 0

        while True:
            # Main cycle in which we yield values to make an iterator

            # Advance iterators when needed / fill caches to keep them long enough
            # Perhaps, some better algorithm is needed to pre-read, this is the simplest

            cache_updated = False
            for i in list(iters.keys()):
                cache = caches[i]
                if len(cache) < cache_size / 2:
                    precache(i, cache_size - len(cache))
                    cache_len = sum(map(len, caches))
                    cache_updated = True

            if cache_updated or (cache_len is not None and (cache_len > order_violation) and (len(docids) < order_violation)):
                while True:
                    max_sum = sum(unread_max)
                    mins = []
                    docids = []
                    maxs = []
                    for w, docid in islice(iter(sorted_mins), order_size):
                        mins.append(-w)
                        docids.append(docid)
                        cacheids = docid2cacheid[docid]
                        # Slower equivalent:
                        # maxs = sum(c.get(docid, m) for c, m in izip(caches, unread_max))
                        # because
                        # -w   = sum(c.get(docid, 0) for c, m in izip(caches, unread_max))
                        maxs.append(max_sum - sum(unread_max[i] for i in cacheids) - w)
                    violated = False
                    for i in xrange(len(mins) - order_violation - 1):
                        # Naive implementation. Can go from tail and do faster
                        # in batches of size of order_violation
                        for m in maxs[i + order_violation:]:
                            if m > mins[i]:
                                violated = True
                                break
                        if violated:
                            break
                    if len(mins) == order_size:
                        # Last check: can the order be violated by out-of-cache elements?
                        if mins[len(mins) - order_violation] < max_sum:
                            violated = True
                    if not violated:
                        break
                    else:
                        # XXX instead we could download enough elements to change
                        # sum(unread_max) by much enough
                        # We need something like "advance_to" method
                        precache(
                                max(enumerate(m / max(l - len(c), 1) for m, l, c in izip(unread_max, lengths, caches)), key=lambda x: x[1])[0],
                                cache_size / 2)
                        cache_len = sum(map(len, caches))

            if not docids:
                break
            minw = mins.pop(0)
            maxw = maxs.pop(0)
            docid = docids.pop(0)
            for i, c in enumerate(caches):
                if docid in c:
                    if c[docid] == unread_max[i]:
                        del c[docid]
                        unread_max[i] = max(six.itervalues(c))
                    else:
                        del c[docid]
                    cache_len -= 1
            used.add(docid)
            sorted_mins.remove((-minw, docid))
            del mins_dict[docid]
            yield docid, (minw + maxw) / 2.0



import itertools as it

import six
from persistent import Persistent
from ZODB.broken import Broken
from zerodbext.catalog.indexes.field import CatalogFieldIndex as _CatalogFieldIndex
from zerodbext.catalog.indexes.common import CatalogIndex
from zerodbext.catalog import RangeValue
from zerodb import trees
from zerodb.catalog.indexes.common import CallableDiscriminatorMixin
from zerodb.util.iter import ListPrefetch

_marker = ()

threshold = 10


def multiunion1(set_type, seqs):
    result = set_type()
    for s in seqs:
        if isinstance(s, six.integer_types):
            s = (s,)
        result.update(s)
    return result


class NewTreeItems(object):
    def __init__(self, items):
        self.items = items
        self.iterator = iter(items)

    def __getitem__(self, i):
        v = self.items.__getitem__(i)
        if isinstance(v, int):
            return (v,)
        else:
            return v

    def __len__(self):
        return self.items.__len__()

    def __iter__(self):
        for nextobj in self.iterator:
            if isinstance(nextobj, int):
                yield (nextobj,)
            else:
                yield nextobj


class CatalogFieldIndex(CallableDiscriminatorMixin, _CatalogFieldIndex):
    family = trees.family32

    # TODO prefetch in search method before multiunion
    # Pass data through apply method

    def __init__(self, discriminator):
        self._init_discriminator(discriminator)
        self._not_indexed = self.family.IF.Set()
        self.clear()

    def applyInRange(self, start, end, excludemin=False, excludemax=False):
        return ListPrefetch(lambda: it.chain.from_iterable(
            ListPrefetch(lambda: NewTreeItems(
                self._fwd_index.values(
                    start, end, excludemin=excludemin, excludemax=excludemax)))))
        # XXX what if these treesets are pretty deep? Need to pre-fetch "first N elements"

    def applyEq(self, value):
        Set = self.family.IF.Set
        docs = self._fwd_index.get(value, None)
        if docs is None:
            return Set()
        elif isinstance(docs, six.integer_types):
            return Set([docs])
        elif isinstance(docs, tuple):
            return Set(docs)
        else:
            return ListPrefetch(lambda: iter(docs))

    def scan_forward(self, docids, limit=None):
        # Batch-prefetch treesets
        # If sorting index is the same as _fwd_index, we already pre-fetched
        # the same objects in the same order!
        fwd_index = ListPrefetch(lambda: self._fwd_index.values())

        n = 0
        for curdocids in fwd_index:
            if isinstance(curdocids, int) and curdocids in docids:
                n += 1
                yield curdocids
                if limit and n >= limit:
                    raise StopIteration
            elif isinstance(curdocids, (tuple, self.family.IF.TreeSet)):
                for docid in curdocids:
                    if docid in docids:
                        n += 1
                        yield docid
                        if limit and n >= limit:
                            raise StopIteration

    def index_doc(self, docid, obj):
        if self.discriminator_callable:
            # Model class definition has a list of virtual fields
            virtuals = getattr(obj.__class__, "_z_virtual_fields", {})
            value = virtuals.get(self.discriminator, _marker)
            if value != _marker:
                try:
                    value = value(obj)
                except:
                    value = _marker
        else:
            value = getattr(obj, self.discriminator, _marker)

        if value is _marker:
            # unindex the previous value
            super(CatalogIndex, self).unindex_doc(docid)

            # Store docid in set of unindexed docids
            self._not_indexed.add(docid)

            return None

        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object %s' %
                             value)

        if isinstance(value, Broken):
            raise ValueError('Catalog cannot index broken object %s' %
                             value)

        if docid in self._not_indexed:
            # Remove from set of unindexed docs if it was in there.
            self._not_indexed.remove(docid)

        return self.inner_index_doc(docid, value)

    def inner_index_doc(self, docid, value):
        """See interface IInjection"""
        rev_index = self._rev_index
        if docid in rev_index:
            docids = self._fwd_index.get(value, ())
            if isinstance(docids, int):
                if docids == docid:
                    return
            elif docid in docids:
                # no need to index the doc, its already up to date
                return
            # unindex doc if present
            self.unindex_doc(docid)

        # Insert into forward index.
        curdocids = self._fwd_index.get(value)
        if curdocids is None:
            self._fwd_index[value] = docid  # integer when only one docid
        else:
            newdocids = curdocids

            if isinstance(curdocids, int):
                newdocids = (curdocids,)
            elif isinstance(curdocids, tuple) and len(curdocids) >= threshold - 1:
                newdocids = self.family.IF.TreeSet(curdocids)
                self._fwd_index[value] = newdocids

            if isinstance(newdocids, tuple):
                newdocids += (docid,)
                self._fwd_index[value] = newdocids
            elif isinstance(newdocids, self.family.IF.TreeSet):
                newdocids.insert(docid)

        # increment doc count
        self._num_docs.change(1)

        # Insert into reverse index.
        rev_index[docid] = value

    def search(self, queries, operator='or'):
        # .apply and .applyAny call this: may need to consider doing
        # an iterative version of this method
        sets = []
        for query in queries:
            if isinstance(query, RangeValue):
                query = query.as_tuple()
            else:
                query = (query, query)
            set = multiunion1(self.family.IF.Set,
                              self._fwd_index.values(*query))
            sets.append(set)

        result = None

        if len(sets) == 1:
            result = sets[0]
        elif operator == 'and':
            sets.sort()
            for set in sets:
                result = self.family.IF.intersection(set, result)
        else:
            result = self.family.IF.multiunion(sets)

        return result

    def unindex_doc(self, docid):
        """See interface IInjection.
        Base class overridden to be able to unindex None values.
        """
        _not_indexed = self._not_indexed
        if docid in _not_indexed:
            _not_indexed.remove(docid)

        rev_index = self._rev_index
        value = rev_index.get(docid, _marker)
        if value is _marker:
            return  # not in index

        del rev_index[docid]

        delvalue = False

        try:
            docids = self._fwd_index[value]
            if isinstance(docids, int):
                delvalue = True
            elif isinstance(docids, tuple):
                newtuple = []
                for i in docids:
                    if i != docid:
                        newtuple.append(i)
                if len(newtuple) == 0:
                    delvalue = True
                else:
                    self._fwd_index[value] = tuple(newtuple)
            elif isinstance(docids, self.family.IF.TreeSet):
                set = self._fwd_index[value]
                set.remove(docid)
                if not set:
                    delvalue = True
        except KeyError:
                pass

        if delvalue:
            del self._fwd_index[value]

        self._num_docs.change(-1)



from BTrees.Interfaces import IIntegerIntegerBTreeModule as IBTreeInterface
import BTrees.IIBTree as TreeModule
from zope.interface import moduleProvides

max_internal_size = 2850
max_leaf_size = 1425


class BTree(TreeModule.BTree):
    max_internal_size = max_internal_size
    max_leaf_size = max_leaf_size


class TreeSet(TreeModule.TreeSet):
    max_internal_size = max_internal_size
    max_leaf_size = max_leaf_size


Set = TreeModule.Set
Bucket = TreeModule.Bucket
difference = TreeModule.difference
union = TreeModule.union
intersection = TreeModule.intersection
multiunion = TreeModule.multiunion
weightedUnion = TreeModule.weightedUnion
weightedIntersection = TreeModule.weightedIntersection

moduleProvides(IBTreeInterface)



from zope.component import getGlobalSiteManager, ComponentLookupError
from zope.interface import implementer
from .interfaces import ICompressor

_gsm = getGlobalSiteManager()


@implementer(ICompressor)
class CommonCompressor(object):
    """
    Common compression class
    One has to set _compress, _decompress and name
    """
    name = b""

    def __init__(self, name=b"", compress=None, decompress=None, args=None, kwargs=None):
        self.name = name
        self._compress = compress
        self._decompress = decompress
        self._signature = self.name and b".c%s$" % self.name
        self._args = args or []
        self._kwargs = kwargs or {}

    def compress(self, data):
        if not data.startswith(self._signature):
            compressed = self._signature + self._compress(data, *self._args, **self._kwargs)
            if len(compressed) < len(data):
                return compressed
        return data

    def decompress(self, data):
        if self._signature and data.startswith(self._signature):
            return self._decompress(data[len(self._signature):])
        else:
            return data

    def register(self, default=False):
        _gsm.registerUtility(self, name=self.name)
        if default:
            _gsm.registerUtility(self)


def decompress(data):
    if data.startswith(b".c"):
        name = data[2:data.find(b"$")].decode()
        return _gsm.getUtility(ICompressor, name).decompress(data)
    else:
        return data


def compress(data):
    try:
        return _gsm.getUtility(ICompressor).compress(data)
    except ComponentLookupError:
        return data



import sys
import os
import platform
import subprocess
import errno
import tempfile

from distutils import ccompiler, log
from setuptools import setup, find_packages

version = '0.99.0a1'

INSTALL_REQUIRES = [
    'BTrees',
    'zope.component>=4.0.0',
    'zodbpickle',
    'ZODB>=5.0.0a6',
    'ZEO>=5.0.0a2',
    'zope.index>=4.0.0',
    'zerodbext.catalog==0.8.4',
    'cachetools',
    'zc.zlibstorage',
    'flask-cors',
    'flask>=0.10',
    'mock',
    'requests>=2.0',
    'jsonpickle',
    'zope.event>=4.0.0',
    'zope.lifecycleevent>=4.0.0',
    'six>=1.7.0',
    'scrypt'
]

TESTS_REQUIRE = [
    'pytest',
    'coverage',
    'path.py',
    'mock',
    'wheel',
    'pytest-cov',
    'pdbpp',
    'zope.testing'
]

entry_points = """
[console_scripts]
zerodb-initdb = zerodb.permissions.base:init_db_script
"""

# The following is to avoid build errors on brand new Amazon Ubuntu
# instances which may not have libffi-dev installed.


# Function copied from cffi 1.5.2
def _ask_pkg_config(resultlist, option, result_prefix='', sysroot=False):
    pkg_config = os.environ.get('PKG_CONFIG', 'pkg-config')
    try:
        p = subprocess.Popen([pkg_config, option, 'libffi'],
                             stdout=subprocess.PIPE)
    except OSError as e:
        if e.errno not in [errno.ENOENT, errno.EACCES]:
            raise
    else:
        t = p.stdout.read().decode().strip()
        p.stdout.close()
        if p.wait() == 0:
            res = t.split()
            res = [x[len(result_prefix):] for x in res
                   if x.startswith(result_prefix)]
            sysroot = sysroot and os.environ.get('PKG_CONFIG_SYSROOT_DIR', '')
            if sysroot:
                # old versions of pkg-config don't support this env var,
                # so here we emulate its effect if needed
                res = [x if x.startswith(sysroot) else sysroot + x for x in res]
            resultlist[:] = res


def can_build_cffi():
    # Windows hopefully grabs binary wheels
    if sys.platform == "win32":
        return True

    # Include dirs copied from cffi 1.5.2
    include_dirs = ["/usr/include/ffi", "/usr/include/libffi"]
    _ask_pkg_config(include_dirs, "--cflags-only-I", "-I", sysroot=True)

    if "freebsd" in sys.platform:
        include_dirs.append("/usr/local/include")

    cc = ccompiler.new_compiler()
    cc.include_dirs = [str(x) for x in include_dirs]  # PY2

    with tempfile.NamedTemporaryFile(mode="wt", suffix=".c") as f:
        f.write('#include "ffi.h"\nvoid f(){}\n')
        f.flush()
        try:
            cc.compile([f.name])
            return True
        except ccompiler.CompileError:
            return False


# If we don't have ffi.h we fall back to pycryptodome.
# Note that the warning is only visible if pip is run with -v.


def have_pycrypto():
    try:
        import Crypto
        return True
    except ImportError:
        return False


def have_pycryptodome():
    try:
        from Crypto.Cipher.AES import MODE_GCM
        return True
    except ImportError:
        return False


def have_aesni():
    if have_pycryptodome():
        from Crypto.Cipher.AES import _raw_aesni_lib
        return _raw_aesni_lib is not None

    else:
        try:
            with open("/proc/cpuinfo", "r") as f:
                info = f.read()
        except IOError:
            info = None

        if (info is None) or ("aes" in info):
            # If we have a platform w/o cpuinfo, assume we have AESNI
            # Perhaps, should call sysctl in OSX
            return True
        else:
            return False


def have_sodium_wheel():
    return ((platform.system() == "Darwin") and
            (platform.mac_ver()[0].startswith("10.10")))

if have_aesni():
    if have_sodium_wheel() or can_build_cffi():
        INSTALL_REQUIRES.append("aes256gcm-nacl")
        if have_pycrypto() and not have_pycryptodome():
            INSTALL_REQUIRES.append("pycrypto")
        else:
            INSTALL_REQUIRES.append("pycryptodome")

    else:
        INSTALL_REQUIRES.append("pycryptodome")
        log.warn(
            "WARNING: ffi.h not found: aes256gcm-nacl optimization disabled")

else:
    INSTALL_REQUIRES.append("pycryptodome")


setup(
    name="zerodb",
    version=version,
    description="End-to-end encrypted database",
    author="ZeroDB Inc.",
    author_email="michael@zerodb.io",
    license="AGPLv3",
    url="http://zerodb.io",
    packages=find_packages(),
    package_data={'zerodb.permissions': ['nobody-key.pem', 'nobody.pem']},
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require={'testing': TESTS_REQUIRE},
    entry_points=entry_points,
)



from zope.index.text import TextIndex as ZopeTextIndex
from zope.index.text.lexicon import CaseNormalizer
from zope.index.text.lexicon import Splitter
from zope.index.text.lexicon import StopWordRemover
from zerodbext.catalog.indexes.text import CatalogTextIndex as _CatalogTextIndex

from zerodb import trees
from zerodb.catalog.indexes.common import CallableDiscriminatorMixin
from zerodb.util.iter import Sliceable

from .text_lexicon import Lexicon
from .text_lucene import IncrementalLuceneIndex
from .text_okapi import OkapiIndex


class CatalogTextIndex(CallableDiscriminatorMixin, _CatalogTextIndex):
    family = trees.family32
    index_class = IncrementalLuceneIndex

    def __init__(self, discriminator, lexicon=None, index=None):
        self._init_discriminator(discriminator)

        self._not_indexed = self.family.IF.Set()

        lexicon = lexicon or Lexicon(Splitter(), CaseNormalizer(), StopWordRemover())
        index = index or self.index_class(lexicon, family=self.family)

        ZopeTextIndex.__init__(self, lexicon, index)
        self.clear()

    def apply(self, querytext, start=0, count=None):
        # For now, let's parse querytext ourselves
        # and later make the queryparser capable to be iterative
        if hasattr(self.index, "_search_all"):
            return Sliceable(lambda: self.index._search_all(querytext))
        else:
            return super(CatalogTextIndex, self).apply(
                    querytext, start=start, count=count)


class CatalogTextIndexOkapi(CatalogTextIndex):
    index_class = OkapiIndex



from logging import getLogger
from threading import local
from transaction.interfaces import ISavepointDataManager
from transaction import get as getTransaction
from zope.interface import implementer

logger = getLogger('collective.indexing.transactions')


class QueueSavepoint(object):
    """ transaction savepoints using the IIndexQueue interface """

    def __init__(self, queue):
        self.queue = queue
        self.state = queue.getState()

    def rollback(self):
        self.queue.setState(self.state)


@implementer(ISavepointDataManager)
class QueueTM(local):
    """ transaction manager hook for the indexing queue """

    def __init__(self, queue):
        local.__init__(self)
        self.registered = False
        self.vote = False
        self.queue = queue

    def register(self):
        if not self.registered:
            try:
                transaction = getTransaction()
                transaction.join(self)
                transaction.addBeforeCommitHook(self.before_commit)
                self.registered = True
            except Exception:
                logger.exception('Exception during register (registered=%s)',
                    self.registered)

    def savepoint(self):
        return QueueSavepoint(self.queue)

    def tpc_begin(self, transaction):
        pass

    def commit(self, transaction):
        pass

    def before_commit(self):
        self.queue.process()
        self.queue.clear()

    def tpc_vote(self, transaction):
        pass

    def tpc_finish(self, transaction):
        self.queue.commit()
        self.registered = False

    def tpc_abort(self, transaction):
        self.queue.abort()
        self.queue.clear()
        self.registered = False

    abort = tpc_abort

    def sortKey(self):
        return str(id(self))



"""Convenience to start ZeroDB servers from Python

Forked :) from ZEO.tests.forker
"""
import logging
import os
import tempfile
import threading
from six.moves.queue import Empty

from ZEO.tests.forker import stop_runner, whine
from ZODB.utils import z64
import ZEO.tests.testssl
import ZODB

from .permissions import subdb, base

logger = logging.getLogger(__name__)

def runner(config, qin, qout, timeout=None,
           debug=False, name=None,
           keep=False, init=True):

    if debug:
        from ZEO.tests.forker import debug_logging
        debug_logging()

    try:

        options = subdb.ZeroDBOptions()
        options.realize(['-C', config])
        server = subdb.ZEOServer(options)
        globals()[(name if name else 'last') + '_server'] = server
        server.open_storages()

        if init:
            password=None
            if isinstance(init, str):
                client_cert = init
            elif isinstance(init, dict):
                client_cert = init.get('cert')
                password = init.get('password')
            else:
                assert init is True
                client_cert = ZEO.tests.testssl.client_cert

            if client_cert:
                with open(client_cert) as f:
                    pem_data = f.read()
            else:
                pem_data = None

            [storage] = server.storages.values()
            base.init_db(storage, 'root', pem_data, False, password=password)

        server.clear_socket()
        server.create_server()
        logger.debug('SERVER CREATED')
        qout.put(server.server.acceptor.addr)
        logger.debug('ADDRESS SENT')
        thread = threading.Thread(
            target=server.server.loop, kwargs=dict(timeout=.2),
            name = None if name is None else name + '-server',
            )
        thread.setDaemon(True)
        thread.start()

        try:
            qin.get(timeout=timeout)
        except Empty:
            pass
        server.server.close()
        thread.join(3)

        if not keep:
            # Try to cleanup storage files
            for storage in server.server.storages.values():
                try:
                    storage.cleanup()
                except AttributeError:
                    pass

        qout.put(thread.is_alive())
        qin.get(timeout=11) # ack
        if hasattr(qout, 'close'):
            qout.close()
            qout.cancel_join_thread()

    except Exception:
        logger.exception("In server thread")

def start_server(storage_conf=None, zeo_conf=None, port=None, keep=False,
                 path='Data.fs', blob_dir=None,
                 suicide=True, debug=False, init=True,
                 threaded=False, start_timeout=33, name=None,
                 ):
    """Start a ZeroDB server in a separate process or thread.

    Takes two positional arguments a string containing the storage conf
    and a ZEOConfig object.

    Returns the ZEO address, the test server address, the pid, and the path
    to the config file.
    """

    if not storage_conf:
        storage_conf = '''
        <filestorage>
          path %s
          pack-gc false
        </filestorage>''' % path

    if blob_dir:
        storage_conf = '<blobstorage>\nblob-dir %s\n%s\n</blobstorage>' % (
            blob_dir, storage_conf)

    if zeo_conf is None or isinstance(zeo_conf, dict):
        if port is None:
            raise AssertionError("The port wasn't specified")

        if isinstance(port, int):
            addr = 'localhost', port
        else:
            addr = port

        from ZEO.tests.forker import ZEOConfig
        z = ZEOConfig(addr)
        if zeo_conf:
            z.__dict__.update(zeo_conf)
        zeo_conf = str(z)

        # jam in SSL configuration:
        zeo_conf = zeo_conf.replace('</zeo>', '''
        <ssl>
           certificate {}
           key {}
           authenticate DYNAMIC
        </ssl>
        </zeo>
        '''.format(ZEO.tests.testssl.server_cert,
                   ZEO.tests.testssl.server_key,
                   ))

    # Store the config info in a temp file.
    tmpfile = tempfile.mktemp(".conf", dir=os.getcwd())
    fp = open(tmpfile, 'w')
    fp.write(str(zeo_conf) + '\n\n')
    fp.write(storage_conf)
    fp.close()

    if threaded:
        from threading import Thread
        from six.moves.queue import Queue
    else:
        from multiprocessing import Process as Thread
        from multiprocessing import Queue

    qin = Queue()
    qout = Queue()
    thread = Thread(
        target=runner,
        args=[tmpfile, qin, qout, 999 if suicide else None],
        kwargs=dict(debug=debug, name=name, keep=keep, init=init),
        name = None if name is None else name + '-server-runner',
        )
    thread.daemon = True
    thread.start()
    try:
        addr = qout.get(timeout=start_timeout)
    except Exception:
        whine("SERVER FAILED TO START")
        if thread.is_alive():
            whine("Server thread/process is still running")
        elif not threaded:
            whine("Exit status", thread.exitcode)
        raise

    def stop(stop_timeout=99):
        stop_runner(thread, tmpfile, qin, qout, stop_timeout)

    return addr, stop
